! $Id: BuildTreeTest.f90 2930 2019-04-27 23:22:26Z ian $
!> @file
!! Defines the BuildTreeTest program.


!*******************************************************************************
!!
!> Test driver of the statement tree construction logic.
!!
!! Breaks the given file into tokens and then folds the tokens into the 
!! statement tree.

PROGRAM BuildTreeTest
  
  USE BaseStNodes
  USE BuildTreePass
  USE Classifier
  USE CmdLine
  USE DefaultHosts
  USE Errors
  USE IncludeLines
  USE LabelsUtils
  USE ScanFreeForm
  USE Scanner
  USE Sources
  USE SourceFiles
  USE Statements
  USE StatementData
  USE StNodes
  USE Strings
  USE SyntaxParts
  
  USE, INTRINSIC :: ISO_FORTRAN_ENV, ONLY: OUTPUT_UNIT, ERROR_UNIT
  
  IMPLICIT NONE
  
  !-----------------------------------------------------------------------------
  ! Locals
  
  ! Format specification for output of an error.
  CHARACTER(*), PARAMETER :: fmt_err = "(A)"
  
  ! Command line options.
  !
  ! In order --ensure-unique, --checks
  TYPE(CmdLineOption) :: cmd_line_options(2)
  
  ! List of non-option arguments.
  TYPE(String), ALLOCATABLE :: arguments(:)
  
  ! Error message
  CHARACTER(:), ALLOCATABLE :: msg
  
  ! Filename read from command line arguments.
  CHARACTER(:), ALLOCATABLE :: filename
  
  INTEGER :: stat             ! Status code for procedure calls.
  TYPE(DefaultHost) :: host   ! Compilation host
  
  ! Local type for a vector of source form objects.
  TYPE form_stack_t
    CLASS(SourceForm), ALLOCATABLE :: item
  END TYPE form_stack_t
  
  ! A stack of source form objects - the first element is for the given 
  ! file and subsequent elements are for INCLUDE'd files.
  TYPE(form_stack_t), ALLOCATABLE :: form_stack(:)
  
  ! Error list for procedure calls.
  TYPE(Error), ALLOCATABLE :: err_list(:)
  
  INTEGER :: ie               ! Error index.
    
  ! Source object for line number information.
  CLASS(Source), POINTER :: source_file
  
  INTEGER :: i                ! Part index.
  
  ! List of tokens read for each statement.
  TYPE(Token), ALLOCATABLE :: tlist(:)
  
  ! Label of the current statement.
  TYPE(Label), ALLOCATABLE :: statement_label
  
  INTEGER :: ist              ! Classification of the current statement.
  
  ! Root of the statement tree.
  TYPE(BaseStNode), TARGET :: base
  
  ! Current tip of the statement tree.
  CLASS(StNode), POINTER :: tip
  
  ! Current part stack.
  INTEGER, ALLOCATABLE :: part_stack(:)
  
  ! Flag to indicate that the statements' worth of tokens was an include line.
  LOGICAL :: is_include
  
  ! Saved copy of the current statements tokens.
  TYPE(Token), ALLOCATABLE :: tmp_tlist(:)
  
  ! Pointer to the parent source object of an INCLUDE'd file.
  CLASS(Source), POINTER :: parent_source
  
  ! Pointer to the source object for the included file.
  CLASS(Source), POINTER :: inc_source
  
  !*****************************************************************************
  
  !-----------------------------------------------------------------------------
  ! Parse the command line and get the filename.
  
  cmd_line_options(1) = CmdLineOption(  &
      'ensure-unique',  &
      DESCRIPTION='Set the ensure unique flag in the top level node.' )
  
  cmd_line_options(2) = CmdLineOption(  &
      'checks',  &
      DESCRIPTION='Carry out additional language related &
        &checks on the tree.' )
  
  CALL ParseCmdLine(cmd_line_options, arguments, stat, msg)
  
  IF (stat /= 0) THEN
    WRITE (ERROR_UNIT, "('Error parsing command line:',A)") msg
    STOP 3
  END IF
  
  IF (SIZE(arguments) == 0) STOP 'Missing filename.'
  IF (SIZE(arguments) > 1)  STOP 'Too many filenames.'
  
  filename = arguments(1)%item
  
  !-----------------------------------------------------------------------------
  ! Create a source form attached to the file.
  
  ALLOCATE(form_stack(1))
  CALL host%OpenSourceForm(filename, form_stack(1)%item, err_list)
  DO ie = 1, SIZE(err_list)
    PRINT fmt_err, .Describe. err_list(ie)
  END DO
  IF (Failure(err_list)) STOP
  
  !-----------------------------------------------------------------------------
  ! Statement read looop
  
  base%ensure_unique = cmd_line_options(1)%PresentCount /= 0
  tip => base
  part_stack = GetParts(tip)
  
  DO WHILE (SIZE(form_stack) > 0)
    
    !---------------------------------------------------------------------------
    ! Read the statement
    
    CALL form_stack(SIZE(form_stack))%item%GetStatement(  &
        part_stack,  &
        err_list,  &
        tlist,  &
        statement_label )
    DO ie = 1, SIZE(err_list)
      PRINT fmt_err, .Describe. (err_list(ie))
    END DO
    IF (Failure(err_list)) STOP 1
    
    ! Zero tokens means end of file.
    IF (SIZE(tlist) == 0) THEN
      ! Close the file.
      CALL form_stack(SIZE(form_stack))%item%GetSource(inc_source)
      CALL inc_source%Reset()
      CALL shrink_stack
      CYCLE
    END IF
    
    ! Save the token list for later output.
    tmp_tlist = tlist
    
    !---------------------------------------------------------------------------
    ! Include line processing.
    
    CALL CheckInclude(tlist, statement_label, is_include, err_list)
    DO ie = 1, SIZE(err_list)
      PRINT fmt_err, .Describe. (err_list(ie))
    END DO
    IF (Failure(err_list)) STOP 1
    
    IF (is_include) THEN
      
      !-----------------------------------------------------------------------
      ! Process an include line by pushing a new source form to the top of
      ! the stack.
      !
      ! Note that because is_include is set, stmt is not defined.
      
      CALL form_stack(SIZE(form_stack))%item%GetSource(parent_source)
      CALL grow_stack
      CALL host%OpenIncludeFile(  &
          tlist(2),  &
          parent_source,  &
          inc_source,  &
          err_list )
      DO ie = 1, SIZE(err_list)
        PRINT fmt_err, .Describe. (err_list(ie))
      END DO
      IF (Failure(err_list)) STOP 1
      
      CALL form_stack(SIZE(form_stack) - 1)%item%Clone(  &
          inc_source,  &
          form_stack(SIZE(form_stack))%item )
      CYCLE
      
    END IF
    
    !---------------------------------------------------------------------------
    ! Statement classification
    
    CALL Classify(tlist, part_stack, ist)
    IF (ist == istUnclassified) THEN
      CALL form_stack(SIZE(form_stack))%item%GetSource(source_file)
      PRINT "(I0,T10,10(A,:,1X))",  &
          source_file%line,  &
          (QueryRaw(tlist(i)), i = 1, MIN(SIZE(tlist), 10))
      
      PRINT *, 'Unclassifiable statement.'
      STOP 1
    END IF
    
    !---------------------------------------------------------------------------
    ! Build the statement tree.
    
    CALL BuildTree(tip, ist, statement_label, tlist, err_list)
    DO ie = 1, SIZE(err_list)
      PRINT fmt_err, .Describe. (err_list(ie))
    END DO
    IF (Failure(err_list)) STOP 2
    
    ! Update the part stack.
    part_stack = GetParts(tip)
  END DO
  
  ! We've run out of source.  Terminate whatever is the current tip.
  CALL TerminateTree(base, tip, err_list)
  DO ie = 1, SIZE(err_list)
    PRINT fmt_err, .Describe. (err_list(ie))
  END DO
  IF (Failure(err_list)) STOP 2
  
  ! Carry out additional checking if asked.
  IF (cmd_line_options(2)%PresentCount /= 0) THEN
    CALL DoBuildTreePassCheck(base, err_list)
    DO ie = 1, SIZE(err_list)
      PRINT fmt_err, .Describe. (err_list(ie))
    END DO
    IF (Failure(err_list)) STOP 2
  END IF
  
  ! Report the statement tree.
  CALL base%DumpTree(OUTPUT_UNIT, '  ')
  
CONTAINS
  
  !***************************************************************************
  !
  ! Grow the stack of source form objects, presumably in response to 
  ! an INCLUDE line.
  !
  ! Accesses the form_stack variable by host association.
  
  SUBROUTINE grow_stack
    ! Temporary for growing the existing form stack.
    TYPE(form_stack_t), ALLOCATABLE :: tmp(:)
    INTEGER :: i            ! Form stack index.
    !*************************************************************************
    ALLOCATE(tmp(SIZE(form_stack) + 1))
    DO i = 1, SIZE(form_stack)
      CALL MOVE_ALLOC(form_stack(i)%item, tmp(i)%item)
    END DO
    CALL MOVE_ALLOC(tmp, form_stack)
  END SUBROUTINE grow_stack
  
  !***************************************************************************
  !
  ! Shrink the stack of source form objects, presumably as processing of 
  ! an INCLUDE'd file completes.
  !
  ! Accesses the form_stack variable by host association.
  
  SUBROUTINE shrink_stack
    ! Temporary for growing the existing form stack.
    TYPE(form_stack_t), ALLOCATABLE :: tmp(:)
    INTEGER :: i            ! Form stack index.
    !*************************************************************************
    ALLOCATE(tmp(SIZE(form_stack) - 1))
    DO i = 1, SIZE(tmp)
      CALL MOVE_ALLOC(form_stack(i)%item, tmp(i)%item)
    END DO
    CALL MOVE_ALLOC(tmp, form_stack)
  END SUBROUTINE shrink_stack
  
END PROGRAM BuildTreeTest
